use std::{
    collections::HashMap,
    convert::TryFrom,
    hash::Hash as Hashable,
    fmt::{self, Display, Formatter},
};
use serde::{Serialize, Deserialize};
use serde_big_array::BigArray;

#[cfg(test)]
mod serde_tests;

/// A Block tagged with its version number.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
enum VersionedBlock {
    V0(Block)
}

/// A container which binds together ciphertext along with the metadata needed to identify,
/// verify and decrypt it.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct Block {
    /// Identifies this block and defines its location in the tree.
    path: Path,
    /// This field contains a collection of `ReadCap`s indexed by the principal who holds them.
    /// `ReadCap`s are envelopments of the key used to encrypt this block.
    read_caps: HashMap<Principal, ReadCap>,
    /// This field is used to verify that the signer of this block had permission to write it.
    /// It contains a certificate chain that must lead back to the root key for the tree this block
    /// is part of.
    write_cap: WriteCap,
    /// The encrypted data contained in this block.
    body: Ciphertext<Vec<u8>>,
    /// The contents of the block are covered by a digital signature contained in this field.  
    signature: Signature
}

/// An envelopment of a key, which is tagged with the principal who the key is meant for.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct ReadCap {
    /// The principal this `ReadCap` was issued to.
    issued_to: Principal,
    /// An encipherment of a block key using the public key of the principal.
    key: Ciphertext<Key>,
}

/// Verifies that a principal is authorized to write blocks in a tree.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct WriteCap {
    /// The principal this `WriteCap` was issued to.
    issued_to: Principal,
    /// The principal that issued this write cap.
    issued_by: Principal,
    /// The path where this write caps's validity begins.
    path: Path,
    /// The point in time after which this write cap is no longer valid.
    expires: Epoch,
    /// A digital signature which covers all of the fields in the write cap except for next.
    signature: Signature,
    /// The next write cap in the chain leading back to the root.
    next: Box<Option<WriteCap>>,
}

/// Fragments are created from blocks using Erasure Encoding and stored with other nodes in the
/// network to provide availability and redundancy of data.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct Fragment {
    /// The path to the block this fragment is from.
    path: Path,
    /// The serial number of this fragment.
    serial: FragmentSerial,
    /// The actual data.
    body: Vec<u8>,
}

/// The body of every non-leaf node in a tree contains this data structure.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct Directory {
    /// The nodes that are attached to the tree at this block.
    nodes: Vec<Principal>,
    /// This block's descendants.
    children: HashMap<String, HashMap<FragmentSerial, FragmentRecord>>,
}

/// Keeps track of which principal is storing a fragment.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct FragmentRecord {
    /// The fragment serial number this record is for.
    serial: FragmentSerial,
    /// The principal who is storing this fragment.
    stored_by: Principal,
}

impl FragmentRecord {
    /// Creates a new `FragmentRecord` whose `serial` and `stored_by` fields are set to
    /// the given values.
    #[allow(dead_code)]
    fn new(serial: u32, stored_by: Hash) -> FragmentRecord {
        FragmentRecord {
            serial: FragmentSerial(serial),
            stored_by: Principal(stored_by),
        }
    }
}

/// An identifier for a security principal, which is any entity that can be authenticated.
#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Hashable)]
struct Principal(Hash);

/// Encrypted data.
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct Ciphertext<T>(T);

/// An identifier for a block in a tree.
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct Path(Vec<String>);

impl Path {
    /// The character that is used to separate path components.
    const SEP: char = '/';
    /// The limit, in bytes, of a `Path`'s length.
    const BYTE_LIMIT: usize = 4096;
}

/// Errors which can occur when converting a string to a `Path`.
#[derive(Debug, PartialEq)]
enum PathError {
    /// Occurs when the number of bytes in a string is greater than `Path::BYTE_LIMIT`.
    PathTooLong(usize),
    /// Indicates that a path string was empty.
    Empty,
    /// Occurs when a component in a path string was empty.
    EmptyComponent,
}

impl Display for PathError {
    fn fmt(&self, formatter: &mut Formatter<'_>) -> fmt::Result {
       match self {
           PathError::PathTooLong(length) => formatter.write_fmt(format_args!(
                "path contained {} bytes, which is over the {} byte limit",
                length,
                Path::BYTE_LIMIT)
           ),
           PathError::Empty => formatter.write_str("path was empty"),
           PathError::EmptyComponent => formatter.write_str("component of path was empty"),
       }
    }
}

impl Path {
    fn consume_component<I: Iterator<Item = (usize, char)>>(
        start: usize, first: char, pairs: &mut I
    ) -> Result<usize, PathError> {
        if first == Path::SEP {
            return Err(PathError::EmptyComponent);
        }
        let end;
        let mut last = start;
        loop {
            match pairs.next() {
                Some((index, Path::SEP)) => {
                    end = index;
                    break;
                },
                Some((index, _)) => last = index,
                None => {
                    end = last + 1;
                    break
                }
            }
        }
        if end == start {
            Err(PathError::EmptyComponent)
        }
        else {
            Ok(end)
        }
    }

    fn trim_leading_sep(mut string: &str) -> Result<&str, PathError> {
        let mut pairs = string.char_indices();
        string = match pairs.next() {
            Some((_, Path::SEP)) => {
                match pairs.next() {
                    Some((index, _)) => &string[index..],
                    None => return Err(PathError::Empty)
                }
            },
            Some((_, _)) => string,
            None => return Err(PathError::Empty)
        };
        Ok(string)
    }
}

impl<'s> TryFrom<&'s str> for Path {
    type Error = PathError;

    fn try_from(mut string: &'s str) -> Result<Path, PathError> {
        {
            let len = string.len();
            if len >= Path::BYTE_LIMIT {
                return Err(PathError::PathTooLong(len))
            }
        }
        string = Path::trim_leading_sep(string)?;
        let mut pairs = string.char_indices();
        let mut components = Vec::new();
        loop {
            let (start, end) = match pairs.next() {
                Some((start, c)) => (start, Path::consume_component(start, c, &mut pairs)),
                None => break
            };
            let slice = &string[start..end?];
            components.push(slice.to_string());
        }
        Ok(Path(components))
    }
}

/// An instant in time represented by the number of seconds since January 1st 1970, 00:00:00 UTC.
#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct Epoch(u64);

/// The serial number of a block fragment.
#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Hashable)]
struct FragmentSerial(u32);

/// A cryptographic hash.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Hashable)]
enum Hash {
    Sha2_256([u8; 32]),
    #[serde(with = "BigArray")]
    Sha2_512([u8; 64]),
}

/// A cryptographic signature.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
enum Signature {
    #[serde(with = "BigArray")]
    Ed25519([u8; 64]),
}

/// A cryptographic key.
#[allow(dead_code)]
#[derive(Debug, PartialEq, Serialize, Deserialize)]
enum Key {
    Xsalsa20Poly1305([u8; 32]),
}

fn main() {
    println!("Hello, world!");
}

#[cfg(test)]
mod tests {
    use super::*;

    fn path_from_str_test_case(
        expected: Result<Path, PathError>, input: &str
    ) -> Result<(), PathError> {
        let result = Path::try_from(input);
        assert_eq!(expected, result);
        Ok(())
    }

    #[test]
    fn path_from_str_multiple_components_ok() -> Result<(), PathError> {
        let expected = Path(vec!["red".to_string(), "green".to_string(), "blue".to_string()]);
        path_from_str_test_case(Ok(expected), "red/green/blue")?;
        Ok(())
    }

    #[test]
    fn path_from_str_one_component_ok() -> Result<(), PathError> {
        let expected = Path(vec!["red".to_string()]);
        path_from_str_test_case(Ok(expected), "red")?;
        Ok(())
    }

    #[test]
    fn path_from_str_leading_slash_ok() -> Result<(), PathError> {
        let expected = Path(vec![
            "orange".to_string(), "banana".to_string(), "shotgun".to_string()
        ]);
        path_from_str_test_case(Ok(expected), "/orange/banana/shotgun")?;
        Ok(())
    }

    #[test]
    fn path_from_str_path_too_long_fail() -> Result<(), PathError> {
        let expected = Err(PathError::PathTooLong(4097));
        let input = "*".repeat(4097);
        path_from_str_test_case(expected, input.as_str())?;
        Ok(())
    }

    #[test]
    fn path_from_str_multiple_slashes_fail() -> Result<(), PathError> {
        let expected = Err(PathError::EmptyComponent);
        path_from_str_test_case(expected, "//orange")?;
        Ok(())
    }
}